/* Name: main.c
 * vim: set sw=4 ts=4 si :
 * Project: anyhkbd
 * Author: Ronald Teune
 * Homepage: http://www.ronaldteune.nl/projects
 * Based on HIDKeys software from http://www.obdev.at/products/avrusb/index.html and avr-usb-slideshow-presenter; http://tuxgraphics.org/electronics/200810/
 * 
 * License: GNU GPL v2 
 */

#define F_CPU   16000000L
#include <util/delay.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <avr/wdt.h>

#include "usbdrv.h"
//#include "oddebug.h"


static void hardwareInit(void)
{
	PORTB = 0x00;   /* activate all pull-ups */
	DDRB = (1 << 0);       /* all pins input, except PB0 */
	PORTC = 0xff;
	DDRC = 0;
	PORTD = 0xff & ~USB_CFG_DMINUS_BIT & ~USB_CFG_DPLUS_BIT;   /* activate pull-ups except on USB lines */
	DDRD = (1 << USB_CFG_DMINUS_BIT) | (1 << USB_CFG_DPLUS_BIT);    /* all pins input except USB (-> USB reset) */
	_delay_ms(10); 
	/* USB Reset by device only required on Watchdog Reset */
	/* delay >10ms for USB reset */
	DDRD = USB_CFG_DPLUS_BIT;    /* 0000 0010 bin: remove USB reset condition */
}

/* ------------------------------------------------------------------------- */

#define NUM_KEYS    3

/* The following function returns an index for the first key pressed. It
 * returns 0 if no key is pressed.
 */
static uchar    keyPressed(void)
{
	if (bit_is_clear(PINC,PINC5)){
		return (1);
	}
	if (bit_is_clear(PINC,PINC2)){
		return (2);
	}
	if (bit_is_clear(PINC,PINC1)){
		return (3);
	}
	return(0);
}


/* ------------------------------------------------------------------------- */
/* ----------------------------- USB interface ----------------------------- */
/* ------------------------------------------------------------------------- */

static uchar    reportBuffer[2];    /* buffer for HID reports */
static uchar    idleRate;           /* in 4 ms units */

PROGMEM const char usbHidReportDescriptor[35] = { /* USB report descriptor */
	0x05, 0x01,                    // USAGE_PAGE (Generic Desktop)
	0x09, 0x06,                    // USAGE (Keyboard)
	0xa1, 0x01,                    // COLLECTION (Application)
	0x05, 0x07,                    //   USAGE_PAGE (Keyboard)
	0x19, 0xe0,                    //   USAGE_MINIMUM (Keyboard LeftControl)
	0x29, 0xe7,                    //   USAGE_MAXIMUM (Keyboard Right GUI)
	0x15, 0x00,                    //   LOGICAL_MINIMUM (0)
	0x25, 0x01,                    //   LOGICAL_MAXIMUM (1)
	0x75, 0x01,                    //   REPORT_SIZE (1)
	0x95, 0x08,                    //   REPORT_COUNT (8)
	0x81, 0x02,                    //   INPUT (Data,Var,Abs)
	0x95, 0x01,                    //   REPORT_COUNT (1)
	0x75, 0x08,                    //   REPORT_SIZE (8)
	0x25, 0x65,                    //   LOGICAL_MAXIMUM (101)
	0x19, 0x00,                    //   USAGE_MINIMUM (Reserved (no event indicated))
	0x29, 0x65,                    //   USAGE_MAXIMUM (Keyboard Application)
	0x81, 0x00,                    //   INPUT (Data,Ary,Abs)
	0xc0                           // END_COLLECTION
};
/* We use a simplifed keyboard report descriptor which does not support the
 * boot protocol. We don't allow setting status LEDs and we only allow one
 * simultaneous key press (except modifiers). We can therefore use short
 * 2 byte input reports.
 * The report descriptor has been created with usb.org's "HID Descriptor Tool"
 * which can be downloaded from http://www.usb.org/developers/hidpage/.
 * Redundant entries (such as LOGICAL_MINIMUM and USAGE_PAGE) have been omitted
 * for the second INPUT item.
 */

/* Keyboard usage values, see usb.org's HID-usage-tables document, chapter
 * 10 Keyboard/Keypad Page for more codes.
 */
#define MOD_CONTROL_LEFT    (1<<0)
#define MOD_SHIFT_LEFT      (1<<1)
#define MOD_ALT_LEFT        (1<<2)
#define MOD_GUI_LEFT        (1<<3)
#define MOD_CONTROL_RIGHT   (1<<4)
#define MOD_SHIFT_RIGHT     (1<<5)
#define MOD_ALT_RIGHT       (1<<6)
#define MOD_GUI_RIGHT       (1<<7)

#define KEY_J       0x0d
#define KEY_M       0x10
#define KEY_K       0x0e

static const uchar  keyReport[NUM_KEYS + 1][2] PROGMEM = {
	/* none */  {0, 0},                     /* no key pressed */
	/*  1 */    {0, KEY_J},
	/*  2 */    {0, KEY_M},
	/*  3 */    {0, KEY_K},
};

static void buildReport(uchar key)
{
	/* This (not so elegant) cast saves us 10 bytes of program memory */
	*(int *)reportBuffer = pgm_read_word(keyReport[key]);
}

uchar    usbFunctionSetup(uchar data[8])
{
	usbRequest_t    *rq = (void *)data;
	
	usbMsgPtr = reportBuffer;
	if((rq->bmRequestType & USBRQ_TYPE_MASK) == USBRQ_TYPE_CLASS){    /* class request type */
		if(rq->bRequest == USBRQ_HID_GET_REPORT){  /* wValue: ReportType (highbyte), ReportID (lowbyte) */
			/* we only have one report type, so don't look at wValue */
			buildReport(keyPressed());
			return sizeof(reportBuffer);
		}else if(rq->bRequest == USBRQ_HID_GET_IDLE){
			usbMsgPtr = &idleRate;
			return 1;
		}else if(rq->bRequest == USBRQ_HID_SET_IDLE){
			idleRate = rq->wValue.bytes[1];
		}
	}else{
		/* no vendor specific requests implemented */
	}
	return 0;
}

/* ------------------------------------------------------------------------- */

int main(void)
{
	uchar   key, lastKey = 0, keyDidChange = 0;
	uint16_t i = 0;
	
	wdt_enable(WDTO_2S);
	hardwareInit();
	usbDeviceDisconnect();
	while(i++ < 255){         // fake USB disconnect for > 250 ms
		wdt_reset();    // if watchdog is active, reset it
		_delay_ms(1);   // library call -- has limited range
	}
	usbDeviceConnect();
	//odDebugInit();
	usbInit();
	sei();
	//DBG1(0x00, 0, 0);
	for(;;){    /* main event loop */
		wdt_reset();
		usbPoll();
		key = keyPressed();
		i++;
		if (i>4000){
			i=0;
		}
		if(lastKey != key && i==0){
			lastKey = key;
			keyDidChange = 1;
		}
		if(keyDidChange && usbInterruptIsReady()){
			keyDidChange = 0;
			/* use last key and not current key status in order to avoid lost
			 *               changes in key status. */
			buildReport(lastKey);
			usbSetInterrupt(reportBuffer, sizeof(reportBuffer));
		}
	}
	return 0;
}

/* ------------------------------------------------------------------------- */
